from mllm.chat import Chat


def test_cached_chat():
    def get_random_number():
        chat = Chat()
        chat += "Give me a random number from 1 to 10"
        res = chat.complete(cache=True)
        return res
    num_1 = get_random_number()
    num_2 = get_random_number()
    assert num_1 == num_2


from mllm.cache.cache_service import caching

def display_chats(disable=False):
    from mllm.chat import ChatLogger
    return ChatLogger(disable)


def display_embed_search(disable=False):
    from mllm.embedding.vector_store import EmbedSearchLogger
    return EmbedSearchLogger(disable)


def refresh_cache(disable=False):
    return caching.refresh_cache(disable)


import ast
import json

"""
# Parse
"""

class Parse:

    @staticmethod
    def obj(src: str):
        try:
            res = ast.literal_eval(src)
        except:
            raise ValueError(f"Invalid Python code: {src}")
        return res

    @staticmethod
    def dict(src: str):
        # find first {
        start = src.find("{")
        # find last }
        end = src.rfind("}")
        if start == -1 or end == -1:
            raise ValueError(f"Invalid json: {src}")
        try:
            # res = ast.literal_eval(src[start:end + 1])
            res = json.loads(src[start:end + 1])
        except:
            raise ValueError(f"Invalid json: {src}")
        return res

    @staticmethod
    def list(src):
        # find first [
        start = src.find("[")
        # find last ]
        end = src.rfind("]")
        if start == -1 or end == -1:
            raise ValueError(f"Invalid json: {src}")
        try:
            res = json.loads(src[start:end + 1])
        except:
            raise ValueError(f"Invalid json: {src}")
        return res

    @staticmethod
    def quotes(src: str):
        """
        Parse a string that is enclosed by quotes ```
        :param src:
        :return: contents of the quoted string
        """
        split_lines = src.split("\n")
        for start, line in enumerate(split_lines):
            if line.startswith("```"):
                title = line[3:]
                break
        else:
            raise ValueError("No closing quotes")
        for end in range(start+1, len(split_lines)):
            if split_lines[end].startswith("```"):
                break
        else:
            raise ValueError("No closing quotes")
        contents = "\n".join(split_lines[start + 1: end])
        return contents

    @staticmethod
    def colon(src: str):
        """
        Parse a string that is separated by colons
        :param src:
        :return: the contents after the colon
        """
        split = src.split(":")
        if len(split) != 2:
            raise ValueError("Invalid colon string")
        contents = split[1].strip()
        return contents



import os

from mllm.setup.inject_env_var import inject_env_var

provider_key_dict = {"openai":["OPENAI_API_KEY"], "anthropic":["ANTHROPIC_API_KEY"], "replicate":["REPLICATE_API_KEY"]}


def setup_wizard():
    # Ask which provider they want to use
    print("Welcome to the MLLM setup wizard!")
    print("This wizard will help you set up your environment to use MLLM.")
    print("Please select which provider you would like to use.")
    for i, provider in enumerate(provider_key_dict.keys()):
        print(f"{i+1}. {provider}")
    provider_choice = input("Enter the number of the provider you would like to use: ")
    provider_choice = int(provider_choice)
    provider_choice -= 1
    provider = list(provider_key_dict.keys())[provider_choice]
    # clear the output
    os.system('cls' if os.name == 'nt' else 'clear')
    print(f"Great! You have selected {provider} as your provider.")


    print(f"You have to set the following environment variables to use {provider}:")
    not_set = False
    for env_var in provider_key_dict[provider]:
        env_var_value = os.environ.get(env_var)
        if env_var_value is None:
            not_set = True
            print(f"{env_var} (current: None)")
        else:
            print(f"{env_var} (current: {env_var_value})")
    if not_set:
        print("Some of the environment variables are not set. Do you want to set them? (y/n)")
    else:
        print("All environment variables are set. Do you want to change them? (y/n)")
    choice = input()
    if choice == "y":
        for env_var in provider_key_dict[provider]:
            env_var_value = input(f"Enter the value for {env_var}: ")
            inject_env_var(env_var, env_var_value)
            os.environ[env_var] = env_var_value
        print("Environment variables set!")
    else:
        print("Environment variables not set.")

    print("Setup wizard ended. The environment variables are:")
    for env_var in provider_key_dict[provider]:
        print(f"{env_var}: {os.environ.get(env_var)}")

    if choice == "y":
        print("Please restart your terminal (IDE) to apply the changes.")

if __name__ == '__main__':
    setup_wizard()



from mllm.utils.maps import parallel_map
import time
def test_parallel_map():
    def wait_for_1_second(x):
        time.sleep(1)
        return x

    for i, res in parallel_map(wait_for_1_second, [1, 2, 3], n_workers=1, title="test"):
        print(res)



import concurrent.futures
import time
from typing import List

from mllm.utils.ipython import is_in_notebook

if is_in_notebook:
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


"""
# Parallel map
"""


default_parallel_map_config = {
    "n_workers": 8
}


def parallel_map(func, *args, n_workers=None, title=None):
    """
    Example usage: `for i, res in parallel_map(lambda x: x + 1, [1, 2, 3, 4, 5], n_workers=4): do_something`
    :param func: The function to apply on each element of args
    :param args: The arguments to apply func
    :param n_workers: Number of workers
    :return:
    """
    # Use concurrent.futures.ThreadPoolExecutor to parallelize
    # Use tqdm to show progress bar
    from mllm.cache.cache_service import caching
    if n_workers is None:
        n_workers = default_parallel_map_config["n_workers"]

    arg_lists = [list(arg) for arg in args]
    start_time = time.time()
    title = title or func.__name__
    with concurrent.futures.ThreadPoolExecutor(max_workers=n_workers) as executor:
        results = []
        for result in tqdm(executor.map(func, *arg_lists, timeout=None), total=len(arg_lists[0]),
                           desc=title):
            results.append(result)
            time_now = time.time()
            if time_now - start_time > 5:
                caching.save()
                start_time = time_now
    caching.save()
    return enumerate(results)


"""
# Nested map
"""


def nested_map(func, nested_list: List):
    """
    Apply func to each element in nested_list and return a nested list with the same structure
    Precondition: list nested can only contain either list or non-list
    :param func:
    :param nested_list:
    :return:
    """
    flat_list = []
    add_to_flat_list(flat_list, nested_list)
    flat_res = func(flat_list)
    nested_res = make_nested_list(flat_res, nested_list)
    return nested_res


def add_to_flat_list(flat_list, nested_list):
    if isinstance(nested_list[0], list):
        for nested_list_ in nested_list:
            add_to_flat_list(flat_list, nested_list_)
    else:
        flat_list.extend(nested_list)


def make_nested_list(flattened_res, nested_list):
    if not isinstance(nested_list[0], list):
        return flattened_res
    nested_res = []
    i = 0
    for nested_list_ in nested_list:
        nested_res.append(
            make_nested_list(flattened_res[i: i + len(nested_list_)], nested_list_))
        i += len(nested_list_)
    return nested_res


def test_nested_map():
    before_map = [[[1], [2]], [3, 4, 5]]
    after_map = nested_map(lambda arr: list(map(lambda x: x + 1, arr)), before_map)
    assert after_map == [[[2], [3]], [4, 5, 6]]



import os


def update_env_var(env_var_name, env_var_value, verbose):
    rc_file = get_rc_file_path()
    # open the rc file
    with open(rc_file, "r") as f:
        lines = f.readlines()
    # find the line to update
    for i, line in enumerate(lines):
        if env_var_name in line:
            lines[i] = f"export {env_var_name}=\"{env_var_value}\"\n"
            break
    else:
        if verbose:
            print(f"Fail to find the environment variable {env_var_name} in the rc file.")
            print("Possible reasons:")
            print("1. The environment variable is not set in the rc file. (we only support RC file)")
            print("2. The environment variable is set in the rc file but with a different format. (we only support export VAR_NAME=\"VAR_VALUE\")")
            print(f"You might need to manually update the environment variable in the rc file.")
        return
    # write the lines back to the rc file
    with open(rc_file, "w") as f:
        f.writelines(lines)
    # source the rc file
    os.system(f"source {rc_file}")
    if verbose:
        print(f"Environment variable {env_var_name} has been updated to {env_var_value}.")
        print(f"RC file: {rc_file} has been updated.")
    return



def inject_env_var(env_var_name, env_var_value, verbose=True):
    """
    Inject an environment variable into the system by writing to the .xxxrc file
    :param env_var_name:
    :param env_var_value:
    :return:
    """
    # judge if the env var is already in the system
    if os.environ.get(env_var_name):
        if verbose:
            print(f"Environment variable {env_var_name} already exists in the system.")
        # compare whether the value is the same
        if os.environ.get(env_var_name) == env_var_value:
            if verbose:
                print(f"Environment variable {env_var_name} has the same value as the one you are trying to inject.")
        else:
            update_env_var(env_var_name, env_var_value, verbose)
        return

    rc_file = get_rc_file_path()

    export_cmd = f"export {env_var_name}=\"{env_var_value}\"\n"
    # open the rc file and write the export command to the first line
    with open(rc_file, "r") as f:
        lines = f.readlines()
    lines.insert(0, export_cmd)
    with open(rc_file, "w") as f:
        f.writelines(lines)

    # source the rc file
    os.system(f"source {rc_file}")

    # return the rc file

    if verbose:
        print(f"Environment variable {env_var_name} with value {env_var_value} has been injected into the system.")
        print(f"RC file: {rc_file} has been updated.")

    return


def get_rc_file_path():
    # get the current terminal type. zsh, bash
    terminal_type = os.environ.get("SHELL")
    terminal_type = terminal_type.split("/")[-1]
    # get the home directory
    home_dir = os.environ.get("HOME")
    # get the rc file
    rc_file = f"{home_dir}/.{terminal_type}rc"
    return rc_file


if __name__ == '__main__':
    inject_env_var("MY_VAR", "MY_VALUE1")
    print(os.environ.get("MY_VAR"))


import os


def check_openai_env():
    return check_env_names(["OPENAI_API_KEY"])

def check_anthropic_env():
    return check_env_names(["ANTHROPIC_API_KEY"])

def check_google_env():
    return True

def check_replicate_env():
    return check_env_names(["REPLICATE_API_KEY"])

def check_env_names(names):
    for name in names:
        if name not in os.environ:
            set_env_var_prompt(name)
            return False
    return True

def set_env_var_prompt(env_var_name):
    print(f"Warning: Please set the environment variable {env_var_name} to the appropriate value.")


from typing import List, Callable

import numpy as np

from mllm.display.show_html import show_json_table
from mllm.embedding.get import get_embeddings
from mllm.utils.logger import Logger


class EmbedSearchLogger(Logger):
    active_loggers = []

    def display_log(self):
        contents = [log for log in self.log_list]
        filenames = [caller_name.split("/")[-1] for caller_name in self.caller_list]
        info_list = []
        for i in range(len(contents)):
            content = contents[i]
            info_list.append({
                "filename": filenames[i],
                "content": content
            })
        show_json_table(info_list)

class VectorStoreItem:
    def __init__(self, item):
        self.vectors = None
        self.srcs = []
        self.weights = []
        self.item = item

    def set_vectors(self, vector_list: List, start):
        self.vectors = np.array(vector_list[start:start + len(self.srcs)])
        return start + len(self.srcs)

    def set_single_vectors(self):
        self.vectors = np.array(get_embeddings(self.srcs))


class VectorStore:
    def __init__(self, score_function=None):
        self.stored_items = []
        self.item_to_index = {}
        if score_function is not None:
            self.score_function = score_function
        else:
            self.score_function = similarity_by_exp

    def add_item(self, item, srcs: List[str], weights: List=None):
        stored_item = VectorStoreItem(item)
        stored_item.srcs = srcs
        if weights is not None:
            stored_item.weights = weights
        else:
            stored_item.weights = np.ones(len(srcs))
        self.stored_items.append(stored_item)
        self.item_to_index[item] = len(self.stored_items) - 1

    def get_top_k_items(self, query: str | List[str], k: int = 10) -> List[str]:
        summed_similarities, items_to_search = self.get_similarities(query, log_stack=2)
        return items_to_search[:k]

    def get_similarities(self, query: str | List[str], items_to_search: List = None, log_stack=1) -> [np.ndarray, List]:
        if isinstance(query, str):
            query = [query]
        query_vecs = np.array(get_embeddings(query))


        if items_to_search is None:
            stored_items = self.stored_items
        else:
            stored_items = [self.stored_items[self.item_to_index[item]] for item in items_to_search]

        no_vector_items = [stored_item for stored_item in stored_items if stored_item.vectors is None]

        srcs = []
        for stored_item in no_vector_items:
            srcs.extend(stored_item.srcs)
        vectors = get_embeddings(srcs)
        start = 0
        for stored_item in no_vector_items:
            start = stored_item.set_vectors(vectors, start)

        inner_prod_list = []
        for stored_item in stored_items:
            inner_prod_list.append(stored_item.vectors.dot(query_vecs.T))

        scores = self.score_function(inner_prod_list)

        # Rank items
        item_rank = np.argsort(scores)[::-1]
        scores = scores[item_rank]
        stored_items = [stored_items[i] for i in item_rank]
        items = [stored_item.item for stored_item in stored_items]
        inner_prod_list = [inner_prod_list[i] for i in item_rank]

        EmbedSearchLogger.add_log_to_all(get_item_similarity_log(stored_items, inner_prod_list, query, top_k=10), stack_depth=log_stack)

        return scores, items

def get_item_similarity_log(stored_items: List[VectorStoreItem], inner_prod_list, query, top_k=10):
    contents = []
    for i, stored_item in enumerate(stored_items):
        inner_prod = inner_prod_list[i]
        max_inner_prod = np.max(inner_prod, 1)
        max_index = np.argmax(max_inner_prod)
        matched_src_inner_prod = inner_prod[max_index]
        matched_query = query[np.argmax(matched_src_inner_prod)]
        content = f"""
Item: {repr(stored_item.item)}
Max Inner Product: {max(matched_src_inner_prod)}
Matched Query: {matched_query}
Matched Src : {stored_item.srcs[max_index]}
"""
        contents.append(content)
    res = "\n".join(contents)
    return res.replace("\n", "<br/>")


def similarity_by_exp(inner_prod_list):
    a = 2.0
    b = 0.1
    # Batch normalization & Add bias
    avg_inner_prod_list = [np.average(inner_prod) for inner_prod in inner_prod_list]
    avg_inner_prod_list = np.array(avg_inner_prod_list)
    average_similarity = np.average(avg_inner_prod_list)
    # Add bias
    inner_prod_list = [inner_prod - average_similarity - b for i, inner_prod in enumerate(inner_prod_list)]
    # Add non-linearity
    inner_prod_list = [np.exp(inner_prod * a) for inner_prod in inner_prod_list]
    # get max inner product
    scores = [np.max(inner_prod) for inner_prod in inner_prod_list]
    return np.array(scores)



def get_vector_store(items: List, get_srcs: Callable[[str], List[str]]) -> VectorStore:
    vector_store = VectorStore()
    src_list = []
    item_list = []
    for i, item in enumerate(items):
        srcs = get_srcs(item)
        src_list.extend(srcs)
        item_list.extend([item] * len(srcs))
        vector_store.add_item(item, srcs)
    return vector_store


def get_vector_store_from_str(str_list: List[str]) -> VectorStore:
    return get_vector_store(str_list, lambda x: [x])


from __future__ import annotations

import inspect
from typing import List


class Logger:
    """
    Please refer to existing children classes if you want to make a new one
    """
    active_loggers: List[Logger] = []

    def __init__(self, disable=False):
        self.log_list = []
        # the path of the file that calls this function
        self.caller_list = []
        self.disable = True if disable != False else False

    def add_log(self, log: any, stack_depth=0):
        caller_path = inspect.stack()[stack_depth + 1].filename
        self.caller_list.append(caller_path)
        self.log_list.append(log)

    @classmethod
    def add_log_to_all(cls, log: any, stack_depth=0):
        for logger in cls.active_loggers:
            logger.add_log(log, stack_depth + 1)

    def __enter__(self):
        if self.disable:
            return self
        # Add self to active_loggers
        self.__class__.active_loggers.append(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return self
        self.__class__.active_loggers.remove(self)
        self.display_log()

    def display_log(self):
        raise NotImplementedError



from mllm import Chat
from mllm.utils.parser import Parse

def test_parse_quotes():
    src = """
123
```python
This is a quoted string
```
123
"""
    res = Parse.quotes(src)
    assert res == "This is a quoted string"

def test_parse_colon():
    src = "title: This is a colon string"
    res = Parse.colon(src)
    assert res == "This is a colon string"

def test_code_gen():
    prompt = """
Generate a code for bubble sort.
Start your answer with ```python
"""
    chat = Chat()
    chat += prompt
    res = chat.complete(parse="quotes", cache=True)
    print(res)

def test_dict_gen():
    prompt = """
Generate a json dict with keys 'a' and 'b' and values 1 and 2
"""
    chat = Chat()
    chat += prompt
    res = chat.complete(parse="dict", cache=True)
    assert res == {"a": 1, "b": 2}



from __future__ import annotations

import inspect
import os
import atexit
import sys

from mllm.cache.cache_embedding import CacheTableEmbed
from mllm.cache.cache_kv import CacheTableKV

def get_main_path():
    return os.path.abspath(sys.argv[0])


def get_cache_path(main_path: str):
    file_name = os.path.basename(main_path)
    dir_name = os.path.dirname(main_path)
    return os.path.join(dir_name, ".llm_cache", file_name + ".json")


class CacheService:

    def __init__(self, base_path=None):
        if base_path is None:
            cache_path = get_cache_path(get_main_path())
        else:
            cache_path = get_cache_path(base_path)
        self.cache_kv: CacheTableKV = CacheTableKV(cache_path)
        self.cache_embed: CacheTableEmbed = CacheTableEmbed(os.path.dirname(cache_path))
        self.cache_kv_other = {self.cache_kv.cache_path: self.cache_kv}
        self.cache_embed_other = {self.cache_kv.cache_path: self.cache_embed}

    def save(self):
        self.cache_kv.save_all_cache_to_file()
        self.cache_embed.save_cache_table()

    def save_used(self):
        n_remove = self.cache_kv.filter_unused_cache()
        if n_remove > 0:
            print(f"Removed {n_remove} unused cache")
        self.cache_kv.save_all_cache_to_file()
        # TODO: save only used embedding cache
        self.cache_embed.save_cache_table()

    def set_main_here(self):
        # get the file path of the caller use inspect
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        main_path = os.path.abspath(module.__file__)
        self._load_cache_on_path(main_path)

    def refresh_cache(self, disable=False):
        """
        Usage: with caching.refresh_cache():
                your codes
        :return: a context manager that refreshes the cache
        """
        return RefreshCacheContext(self.cache_kv, disable=disable)

    def disable_cache(self, disable=False):
        """
        Usage: with caching.disable_cache():
                your codes
        :return:
        """
        return DisableCacheContext(self.cache_kv, disable=disable)

    def _load_cache_on_path(self, main_path):
        cache_path = get_cache_path(main_path)
        cache_dir = os.path.dirname(cache_path)

        if cache_path in self.cache_kv_other:
            self.cache_kv = self.cache_kv_other[cache_path]
        else:
            self.cache_kv = CacheTableKV(cache_path)
            self.cache_kv_other[cache_path] = self.cache_kv

        if cache_dir in self.cache_embed_other:
            self.cache_embed = self.cache_embed_other[cache_dir]
        else:
            self.cache_embed = CacheTableEmbed(cache_dir)
            self.cache_embed_other[cache_dir] = self.cache_embed


class RefreshCacheContext:
    def __init__(self, cache_kv: CacheTableKV, cache_type: str = "", disable=False):
        """
        :param cache_type: The type of cache to refresh. If type is "", then all cache will be
        refreshed
        """
        self.cache_type = cache_type
        self.cache_kv = cache_kv
        self.already_refreshed = False
        self.disable = True if disable != False else False

    def __enter__(self):
        if self.disable:
            return
        if self.cache_type != "":
            if self.cache_type not in self.cache_kv.types_to_refresh:
                self.cache_kv.types_to_refresh.add(self.cache_type)
            else:
                self.already_refreshed = True
        else:
            if self.cache_kv.refresh_all:
                self.already_refreshed = True
            else:
                self.cache_kv.refresh_all = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return
        if self.cache_type != "":
            if not self.already_refreshed:
                self.cache_kv.types_to_refresh.remove(self.cache_type)
        else:
            if not self.already_refreshed:
                self.cache_kv.refresh_all = False
        self.cache_kv.apply_cache_update()


class DisableCacheContext:
    def __init__(self, cache_kv: CacheTableKV, disable=False):
        self.cache_kv: CacheTableKV = cache_kv
        self.disable = True if disable != False else False

    def __enter__(self):
        if self.disable:
            return
        self.cache_kv.inactive = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.disable:
            return
        self.cache_kv.inactive = False


caching = CacheService()

# save cache on exit
atexit.register(caching.save)



import webbrowser
from json2html import json2html
import tempfile

from mllm.utils.ipython import is_in_notebook

display_config = {
    'show_in_notebook': True,
}

def show_html(html: str):
    if is_in_notebook and display_config['show_in_notebook']:
        show_html_in_notebook(html)
    else:
        show_html_in_browser(html)

def show_html_in_browser(html: str):
    with tempfile.NamedTemporaryFile('w', delete=False, suffix='.html') as file:
        file.write(html)
        filename = 'file:///' + file.name
    webbrowser.open_new_tab(filename)

def show_html_in_notebook(html: str):
    from IPython.display import display, HTML
    display(HTML(html))


def show_json_table(json_list):
    """
    Show a json data as a table in browser or notebook
    :param json_list: The json data to show
    :return:
    """
    if len(json_list) == 0:
        print("No data to show in show_json_table(). Skipping...")
        return

    table_html = json2html.convert(json=json_list, escape=False)
    # write the html to a file
    html = f"""
<!DOCTYPE html>
<html>
<head>
    <title>Table</title>
    <style>
        table {{
            border-collapse: collapse;
            width: 100%;
        }}

        th, td {{
            border: 1px solid #dddddd;
            text-align: left;
            padding: 8px;
        }}

        tr:nth-child(even) {{
            background-color: #f2f2f2;
        }}
    </style>
</head>
<body>
    {table_html}
</body> 
</html>
"""
    show_html(html)


if __name__ == '__main__':
    data = [
        {
            "Name": "John",
            "Age": 30,
            "City": "New York"
        },
        {
            "Name": "Peter",
            "Age": 45,
            "City": "Boston"
        }
    ]
    show_json_table(data)



import os

import numpy as np


class CacheTableEmbed:
    def __init__(self, cache_dir: str):
        self.cache_tables = {}
        self.cache_dir = cache_dir

    def get_cache_path(self, model_name: str):
        return os.path.join(self.cache_dir, f"embed_{model_name}.npy")

    def load_cache_table(self, model_name: str):
        if model_name in self.cache_tables:
            return self.cache_tables[model_name]
        embedding_cache_path = self.get_cache_path(model_name)
        if os.path.exists(embedding_cache_path):
            embedding_cache = np.load(embedding_cache_path, allow_pickle=True).item()
        else:
            embedding_cache = {}
        self.cache_tables[model_name] = embedding_cache
        return embedding_cache

    def save_cache_table(self):
        os.makedirs(self.cache_dir, exist_ok=True)
        for model_name in self.cache_tables:
            embedding_cache_path = self.get_cache_path(model_name)
            np.save(embedding_cache_path, self.cache_tables[model_name])

    def clear_cache_table(self, model_name: str):
        if model_name in self.cache_tables:
            self.cache_tables[model_name] = {}



from mllm import get_embeddings, caching
from mllm.config import default_models
from mllm.embedding import get_vector_store_from_str


def test_vector_store_basic():
    vector_store = get_vector_store_from_str(["banana", "headset", "Mike"])
    res = vector_store.get_top_k_items("earphone")
    assert res[0] == "headset"
    res = vector_store.get_top_k_items("apple")
    assert res[0] == "banana"
    res = vector_store.get_top_k_items("Jackson")
    assert res[0] == "Mike"

def test_vector_store_basic2():
    vector_store = get_vector_store_from_str(["banana", "headset", "Mike"])
    res = vector_store.get_top_k_items(["earphone", "PC"])
    assert res[0] == "headset"

def test_lazy_embedding():
    get_embeddings("1")
    caching.cache_embed.clear_cache_table(default_models["embedding"])

    texts_1 = ["banana", "headset", "Mike"]
    texts_2 = ["earphone", "PC"]

    res_1 = get_embeddings(texts_1)
    res_2 = get_embeddings(texts_2)

    print(res_1)
    print(res_2)


from mllm import Chat, display_chats

if __name__ == '__main__':
    with display_chats():
        chat = Chat()
        chat += "Give me a random number from 1 to 10"
        res = chat.complete()
        print(res)

    # put something in the argument to disable the logger
    with display_chats(1):
        chat = Chat()
        chat += "Give me a random number from 1 to 10"
        res = chat.complete()
        print(res)



from mllm import Chat


def test_image_chat():
    chat = Chat()
    chat.add_user_message("What is in this image?")
    chat.add_image_message(
        "EvoLogoOrg.png")
    res = chat.complete()
    print(res)


from mllm import Chat, caching

def get_random_number():
    chat = Chat()
    chat += "Give me a random number from 1 to 100"
    res = chat.complete(cache=True)
    return res

if __name__ == '__main__':
    num_1 = get_random_number()
    print(num_1)
    with caching.refresh_cache():
        num_1 = get_random_number()
    print(num_1)
    num_1 = get_random_number()
    print(num_1)


from mllm.chat import Chat
from mllm.embedding import get_embeddings


def test_basic():
    inputs = ["a", "b", "c"]
    res = get_embeddings(inputs)
    print(res)


def test_chat():
    chat = Chat()
    chat.add_user_message("Who is your developer?")
    res = chat.complete()
    print(res)


def test_chat_plus():
    chat = Chat()
    chat += "Who is your developer?"
    res = chat.complete()
    print(res)


def test_system_message():
    chat = Chat(system_message="You are an assistant who only reply in `yes` or `no`")
    chat += "Are you a human?"
    res = chat.complete()
    print(res)


import mllm
from mllm import Chat
from mllm.embedding import get_vector_store_from_str


def test_image_display():
    with mllm.debug.display_chats():
        chat = Chat()
        chat.add_user_message("What is in this image?")
        chat.add_image_message(
            "EvoLogoOrg.png")
        res = chat.complete(cache=True)

def test_chat_with_html_tags():
    with mllm.debug.display_chats():
        chat = Chat()
        chat.add_user_message("<input>What is your name?</input>")
        res1 = chat.complete()


def test_cached_display():
    with mllm.debug.refresh_cache():
        chat = Chat()
        chat.add_user_message("What is your name?")
        res1 = chat.complete(cache=True)
    with mllm.debug.display_chats():
        chat = Chat()
        chat.add_user_message("What is your name?")
        res2 = chat.complete(cache=True)
        assert res1 == res2

def test_chat_display_with_indent():
    with mllm.debug.display_chats():
        chat = Chat()
        chat.add_user_message("""
Explain the following code
```python
def greet():
    print("Hello")
greet()
```
""")
        res = chat.complete(cache=True)


def test_vector_store_display():
    with mllm.debug.display_embed_search():
        vector_store = get_vector_store_from_str(["banana", "headset", "Mike"])
        res = vector_store.get_top_k_items(["earphone", "PC"])
        assert res[0] == "headset"




from __future__ import annotations

import hashlib
import json
import os
from typing import Dict, Optional, List, Set



def get_hash(data: any, type: str) -> str:
    return hashlib.sha1(json.dumps([data, type]).encode("utf-8")).hexdigest()


class Cache:
    """
    The class for storing cache.
    """

    def __init__(self, value, hash: str, input: any, type: str,
                 meta: Optional[Dict] = None):
        self.value = value
        # self.hash should be the same as get_hash(input, type)
        self.hash: str = hash
        self.input: any = ""#input
        self.type: str = type
        self.meta = meta

    def get_self_dict(self):
        return {
            "value": self.value,
            "hash": self.hash,
            "input": self.input,
            "type": self.type,
            "meta": self.meta
        }

    def set_cache(self, value: any, meta: Optional[Dict] = None):
        self.value = value
        self.meta = meta

    def is_valid(self):
        return self.value is not None


CacheTable = Dict[str, Cache]


def serialize_cache_table(cache_table: Dict[str, Cache]):
    res = []
    for key, cache in cache_table.items():
        res.append(cache.get_self_dict())
    return json.dumps(res, indent=1)


class CacheTableKV:
    def __init__(self, cache_path: str):
        self.cache_path = cache_path
        # Map from hash to cache
        self.cache_table: Dict[str, Cache] = self.load_cache_table()
        # List of pending cache
        self.pending_cache: List[Cache] = []
        # List of active cache. Used for garbage collection
        self.active_cache_hash: Set[str] = set()
        #
        self.types_to_refresh: Set[str] = set()
        #
        self.refresh_all: bool = False
        #
        self.types_used: Set[str] = set()
        #
        self.inactive = False

    def save_all_cache_to_file(self):
        self.apply_cache_update()
        if len(self.cache_table) == 0:
            return
        os.makedirs(os.path.dirname(self.cache_path), exist_ok=True)
        with open(self.cache_path, "w") as f:
            f.write(serialize_cache_table(self.cache_table))

    def filter_unused_cache(self) -> int:
        n_removed_cache = 0
        new_cache_table = {}
        for hash, cache in self.cache_table.items():
            if hash in self.active_cache_hash:
                new_cache_table[hash] = cache
            else:
                n_removed_cache += 1
        self.cache_table = new_cache_table
        return n_removed_cache

    def read_cache(self, input: any, type: str, create_cache=True) -> Cache | None:
        if self.inactive:
            return None
        hash = get_hash(input, type)
        self.types_used.add(type)
        un_hit = hash not in self.cache_table
        if type in self.types_to_refresh or self.refresh_all:
            un_hit = True
        if un_hit:
            if create_cache:
                new_cache = Cache(None, hash, input, type)
                self.add_cache(new_cache)
                return new_cache
            else:
                return None
        cache_hit = self.cache_table[hash]
        self.active_cache_hash.add(hash)
        return cache_hit

    def add_cache(self, cache: Cache):
        self.pending_cache.append(cache)

    def apply_cache_update(self):
        remaining_cache = []
        for cache in self.pending_cache:
            if cache.is_valid():
                self.active_cache_hash.add(cache.hash)
                self.cache_table[cache.hash] = cache
            else:
                remaining_cache.append(cache)
        self.pending_cache = remaining_cache

    def discard_cache_update(self):
        self.pending_cache = []

    def load_cache_table(self) -> CacheTable:
        if not os.path.exists(self.cache_path):
            return {}
        with open(self.cache_path, "r") as f:
            try:
                cache_list = json.load(f)
            except json.decoder.JSONDecodeError:
                cache_list = []
        cache_table = {}
        for cache_dict in cache_list:
            cache = Cache(cache_dict["value"], cache_dict["hash"], cache_dict["input"],
                          cache_dict["type"], cache_dict["meta"])
            cache_table[cache.hash] = cache
        return cache_table







default_models = {
    "normal": "gpt-3.5-turbo",
    "expensive": "gpt-4-turbo-preview",
    "vision": "gpt-4-vision-preview",
    "embedding": "text-embedding-3-large"
}


default_options = {
    "temperature": 0.2,
    "max_tokens": None,
    "frequency_penalty": None,
    "timeout": 600,
    "seed": None,
}


class CacheOptions:
    def __init__(self):
        self.max_embedding_vecs = 1000
        self.max_chat = 1000


from __future__ import annotations

import hashlib
from typing import List, Dict

import numpy as np
from litellm import embedding
from mllm.cache.cache_service import caching
from mllm.config import default_models


class LazyEmbedding:
    lazy_embeddings: Dict[str, List[LazyEmbedding]] = {}

    def __init__(self, src: str, model: str, hash_key, cache_table):
        self.src = src
        self.model = model
        self.hash_key = hash_key
        self.cache_table = cache_table
        self.embedding = None
        if model not in LazyEmbedding.lazy_embeddings:
            LazyEmbedding.lazy_embeddings[model] = [self]
        else:
            LazyEmbedding.lazy_embeddings[model].append(self)


    def __array__(self):
        if self.embedding is None:
            self.flush()
        return self.embedding

    def to_list(self):
        return list(self.__array__())

    def flush(self):
        lazy_embeddings = LazyEmbedding.lazy_embeddings[self.model]
        embeddings = _get_embeddings(self.model, [le.src for le in lazy_embeddings])
        for i, item in enumerate(lazy_embeddings):
            item.embedding = embeddings[i]
            self.cache_table[item.hash_key] = item.embedding
        LazyEmbedding.lazy_embeddings[self.model] = []

    def __str__(self):
        return str(self.__array__())

    def __repr__(self):
        return repr(self.__array__())


def get_embeddings(texts: list[str], model=None, lazy=True) -> list[list[float]]:
    if model is None:
        model = default_models["embedding"]

    cache_table = caching.cache_embed.load_cache_table(model)
    hash_keys = [hashlib.md5(text.encode()).hexdigest() for text in texts]

    embeddings = []
    index_for_eval = []
    texts_without_cache = []
    for i, text in enumerate(texts):
        if len(text) == 0:
            embeddings.append(0)
            continue
        if hash_keys[i] not in cache_table:
            texts_without_cache.append(text)
            embeddings.append(None)
            index_for_eval.append(i)
        else:
            embeddings.append(cache_table[hash_keys[i]])
    if len(texts_without_cache) > 0:
        if not lazy:
            res = _get_embeddings(model, texts_without_cache)
            for i, r in zip(index_for_eval, res):
                cache_table[hash_keys[i]] = r
                embeddings[i] = r
        else:
            for i in index_for_eval:
                le = LazyEmbedding(texts[i], model, hash_keys[i], cache_table)
                embeddings[i] = le

    return embeddings


def _get_embeddings(model, texts_without_cache):
    res = embedding(model, input=texts_without_cache)
    res = [np.array(r['embedding']) for r in res.data]
    return res



from __future__ import annotations

import base64
import copy
import html
from io import BytesIO

import httpx
from PIL.Image import Image
from litellm import completion

from mllm.cache.cache_service import caching
from mllm.utils.logger import Logger
from mllm.config import default_models, default_options
from mllm.display.show_html import show_json_table
from mllm.utils.parser import Parse

n_chat_retry = 3

def encode_image(image_file: BytesIO):
    return base64.b64encode(image_file.read()).decode('utf-8')


def get_chat_in_html(chat: Chat):
    res = []
    message_to_api = chat.get_messages_to_api()
    for entry in message_to_api:
        if isinstance(entry["content"], str):
            entry["content"] = [{"type": "text", "text": entry["content"]}]
        content = []
        for item in entry["content"]:
            if item["type"] == "text":
                text = item["text"]
                text = html.escape(text)
                text = text.replace("\n", "<br/>")
                text = text.replace(" ", "&nbsp;")
                content.append(text)
            elif item["type"] == "image_url":
                content.append("<image src='{}' style='max-height: 200px;'/>".format(item["image_url"]["url"]))
        content = "<br/>".join(content)
        res.append(f"------{entry['role']}------<br/> {content}")
    return "<br/>".join(res)

class ChatLogger(Logger):
    active_loggers = []

    def display_log(self):
        contents = [get_chat_in_html(chat) for chat in self.log_list]
        filenames = [caller_name.split("/")[-1] for caller_name in self.caller_list]
        info_list = []
        for i in range(len(contents)):
            content = contents[i]
            info_list.append({
                "filename": filenames[i],
                "content": content
            })
        show_json_table(info_list)


class ParseError(Exception):
    pass


def parse_res(parse, res):
    if parse not in ["dict", "list", "obj", "quotes", "colon"]:
        raise ValueError("Invalid parse type")
    try:
        res = Parse.__dict__[parse](res)
    except Exception as e:
        raise ParseError(f"Failed to parse the result: {e}")
    return res


class Chat:
    """
    Class for chat completion
    """

    def __init__(self, user_message=None, system_message: any = None):
        self.messages = []
        self.system_message = system_message
        if user_message is not None:
            self._add_message(user_message, "user")

    """
    ## Message editing and output
    """

    def _add_message(self, content: any, role: str):
        self.messages.append({
            "role": role,
            "content": {
                "type": "text",
                "text": content
            }
        })

    def _add_image_message(self, data: str, media_type: str):
        self.messages.append({
            "role": "user",
            "content": {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/{media_type};base64,{data}"
                }
            }
        })

    def add_user_message(self, content: any):
        self._add_message(content, "user")

    def add_image_message(self, image_or_image_path: str | Image, more: dict = None):
        """
        :param image_path: The path of the image
        """

        if isinstance(image_or_image_path, str):
            from_internet = image_or_image_path.startswith(
                "http://") or image_or_image_path.startswith("https://")
            if not from_internet:
                img_io = open(image_or_image_path, "rb")
            else:
                img_io = BytesIO(httpx.get(image_or_image_path).content)
            media_type = image_or_image_path.split(".")[-1]
            media_type = media_type.lower()
            base64_image = encode_image(img_io)
        else:
            img_io = BytesIO()
            image_or_image_path.save(img_io, format="png")
            media_type = "png"
            base64_image = base64.b64encode(img_io.getvalue()).decode('utf-8')

        assert media_type in ["jpg", "jpeg", "png", "gif", "webp"]
        if media_type == "jpg":
            media_type = "jpeg"

        img_io.close()
        self._add_image_message(base64_image, media_type)

    def add_assistant_message(self, content: any):
        self._add_message(content, "assistant")

    def get_messages_to_api(self):
        """
        :return: messages for sending to model APIs
        """
        messages = []
        if self.system_message is not None:
            messages.append({
                "role": "system",
                "content": self.system_message
            })

        if len(self.messages) == 0:
            return messages

        content_list = []
        curr_role = self.messages[0]["role"]
        content_dict = {
            "role": curr_role,
            "content": content_list
        }
        messages.append(content_dict)

        for message in self.messages:
            if message["role"] != curr_role:
                content_list = []
                curr_role = message["role"]
                content_dict = {
                    "role": curr_role,
                    "content": content_list
                }
                messages.append(content_dict)
            content_list.append(message["content"])

        # Merge text messages into one
        for message in messages:
            if message["role"] != "user":
                continue
            content_list = []
            merge_failed = False
            for item in message["content"]:
                if item["type"] == "text":
                    content_list.append(item["text"])
                else:
                    merge_failed = True
                    break
            if not merge_failed:
                message["content"] = "\n\n".join(content_list)

        return messages

    def contains_image(self):
        for message in self.messages:
            content = message["content"]
            if isinstance(content, list):
                for item in content:
                    if isinstance(item, dict) and item["type"] == "image":
                        return True
            elif isinstance(content, dict):
                if "image_url" in content:
                    return True
        return False

    """
    ## Chat completion functions
    """

    def complete(self, model=None, cache=False, expensive=False, parse=None, retry=True, options=None):
        """
        :param model: The name of the model to use. If None, the default model will be used.
        :param cache: Whether to use cache. If True, the result will be cached.
        :param expensive: Whether to use the expensive model. If True, the default expensive model will be used.
        :param parse: How to parse the result. Options: "dict", "list", "obj", "quotes", "colon". See http://mllm.evoevo.org/parsing for details.
        :param retry: Whether to retry if the completion fails.
        :param options: Additional options for the completion model.
        :return: The completion result from the model. If parse is set, a parsed result will be returned.
        """
        if options is None:
            options = {}
        options = {**default_options, **options}
        if model is None:
            if not expensive:
                model = default_models["normal"]
            else:
                model = default_models["expensive"]
            if self.contains_image():
                model = default_models["vision"]
        for n_tries in range(n_chat_retry):
            try:
                res = self._complete_chat_impl(model, cache, options)
                if parse is not None:
                    res = parse_res(parse, res)
                return res
            # Retry if the completion fails on TimeoutError or ParseError
            except (TimeoutError, ParseError) as e:
                if not retry:
                    raise e
                # Disable cache for retry
                cache = False
                print(e)
                # print stack here
                import traceback
                print(traceback.format_exc())
                print("Retrying...")
        raise Exception("Failed to complete chat. Did you set the correct API key? Did you prompt the model to output the expected parsing format?")

    def _complete_chat_impl(self, model: str, use_cache: bool, options):
        messages = self.get_messages_to_api()
        if use_cache:
            cache = caching.cache_kv.read_cache(messages, "chat")
            if cache is not None and cache.is_valid():
                self.add_assistant_message(cache.value)
                ChatLogger.add_log_to_all(self, stack_depth=2)
                return cache.value

        options = options or {}

        res = completion(model, messages=messages, **options).choices[0].message.content

        self.add_assistant_message(res)

        ChatLogger.add_log_to_all(self, stack_depth=2)

        if use_cache and cache is not None:
            cache.set_cache(res)
        return res

    """
    ## Magic methods
    """

    def __str__(self):
        res = []
        message_to_api = self.get_messages_to_api()
        for entry in message_to_api:
            if isinstance(entry["content"], str):
                content = entry["content"]
            else:
                content = []
                for item in entry["content"]:
                    if item["type"] == "text":
                        content.append(item["text"])
                    elif item["type"] == "image_url":
                        content.append("<image/>")
                content = "\n".join(content)
            res.append(f"------{entry['role']}------\n {content}")
        return "\n".join(res)

    def __repr__(self):
        return f"<{self.__class__.__name__}> {self.system_message!r}"

    def __copy__(self):
        new_chat = Chat(system_message=self.system_message)
        new_chat.messages = copy.deepcopy(self.messages)
        return new_chat

    def __add__(self, other):
        assert isinstance(other, str)
        self.add_user_message(other)
        return self







"""
# IPython
"""

def check_if_in_notebook() -> bool:
    try:
        from IPython import get_ipython
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False

is_in_notebook = check_if_in_notebook()


from tenacity import retry, wait_fixed, stop_after_attempt, retry_if_exception

standard_multi_attempts = retry(
    wait=wait_fixed(0.5),
    stop=(stop_after_attempt(3)),
    retry=retry_if_exception(lambda e: True),
    reraise=False,
)


def test_standard_multi_attempts():
    @standard_multi_attempts
    def a():
        print("a")
        raise ValueError("a")
    a()



import os

from mllm import Chat


def test_no_api_key():
    os.environ["OPENAI_API_KEY"] = ""
    chat = Chat()
    chat.add_user_message("What is your name?")
    res = chat.complete()

def test_failed_parsing():
    chat = Chat()
    chat.add_user_message("What is your name?")
    res = chat.complete(parse="dict")


from mllm.config import default_models
from mllm.env_utils import check_replicate_env, check_anthropic_env, check_openai_env

def set_default_to_google():
    default_models["normal"] = "gemini-1.0-pro"
    default_models["expensive"] = "gemini-1.0-pro"
    default_models["vision"] = "gemini-1.0-pro-vision"
    default_models["embedding"] = "vertex_ai/textembedding-gecko"

def set_default_to_openai(use_gpt_4=False):
    check_openai_env()
    default_models["normal"] = "gpt-3.5-turbo"
    default_models["expensive"] = "gpt-4-turbo-preview" if not use_gpt_4 else "gpt-4"
    default_models["vision"] = "gpt-4-vision-preview"
    default_models["embedding"] = "text-embedding-3-large"

def set_default_to_anthropic(expensive_vision_model=False):
    check_anthropic_env()
    default_models["normal"] = "claude-3-sonnet-20240229"
    default_models["expensive"] = "claude-3-opus-20240229"
    default_models["vision"] = "claude-3-sonnet-20240229" if not expensive_vision_model else "claude-3-opus-20240229"
    default_models["embedding"] = "text-embedding-3-large"

def set_default_to_llama():
    check_replicate_env()
    default_models["normal"] = "replicate/meta/llama-2-13b-chat"
    default_models["expensive"] = "replicate/meta/llama-2-70b-chat"
    default_models["vision"] = "gpt-4-vision-preview"
    default_models["embedding"] = "text-embedding-3-large"